with Strings, Ints, Globals;
use Strings, Ints, Globals;

pragma Elaborate_All (Strings,Ints,Globals);

package body Hopf is

  function Need2(M: F3Mode) return Boolean is
  begin
    return not (M.I=0 or else M.F2Err);
  end Need2;

  procedure Norm(M: in F3Mode; S: in out Scalar) is
  begin
    if M.I=0 or else M.F2Err then
      Copy(SOne,S);       -- error modes have norm 1
    else
      Assign(Up_RhoNN,S); -- coeff
    end if;
  end Norm;

  procedure MaxNorm(M: in F3Mode; S: out Radius) is
  begin
    if M.I=0 or else M.F2Err then
      S := One;      -- error modes have norm 1
    else
      S := Up_RhoNN; -- coeff
    end if;
  end MaxNorm;

  procedure Assign(M: in F3Mode; Q: in out Fourier3) is
    --- assuming that M is fine
    I: Indx    renames M.I;
    J: FreqE1  renames M.J;
    K: FreqE2  renames M.K;
  begin
    SetZero(Q);
    if M.F2Err then
      Q.E(J,K)(I) := OneMinusTiny;
    else
      Copy(SOne,Q.C(J,K)(I));
    end if;
  end Assign;

  procedure Assign2(M: in F3Mode; Q: in out Fourier3) is
    --- assuming that M is fine
    I: Indx    renames M.I;
    J: FreqE1  renames M.J;
    K: FreqE2  renames M.K;
  begin
    SetZero(Q);
    if M.I<=0 or else M.F2Err then raise Inconsistent_Data; end if;
    Copy(SOne,Q.C(J,K)(-I));
  end Assign2;

  procedure AddProd(M: in F3Mode; S: in Scalar; Q: in out Fourier3; Check: in Boolean) is
    I: Indx   renames M.I;
    J: FreqE1 renames M.J;
    K: FreqE2 renames M.K;
    pragma Unreferenced(Check);
  begin
    if not IsZero(S) then
      if M.F2Err then
        raise Not_Implemented;
      else
        Add(S,Q.C(J,K)(I));
      end if;
    end if;
  end AddProd;

  procedure Extract_Coeff(M: in F3Mode; Q: in out Fourier3; S: in out Scalar) is
    --- assuming M.F2Err=False
    --- S encloses the coefficient
    I: Indx   renames M.I;
    J: FreqE1 renames M.J;
    K: FreqE2 renames M.K;
  begin
    if M.Simple then
      Copy(Q.C(J,K)(I),S);
      --- should set some things to zero, but not needed here
    else
      raise Not_Implemented;
    end if;
  end Extract_Coeff;

  procedure Fail(M: in F3Mode; Q: in out Fourier3; S: in out Scalar) is
    pragma Unreferenced(M);
    pragma Unreferenced(Q);
    pragma Unreferenced(S);
  begin
    raise Not_Implemented;
  end Fail;

  procedure Extract(M: in F3Mode; Q: in out Fourier3; S: in out Scalar; Check: in Boolean) is
    --- assuming M is consistent
    pragma Unreferenced(Check);
  begin
    if M.F2Err then
      Extract_Error(M,Q,S); --- Fail
    else
      Extract_Coeff(M,Q,S);
    end if;
  end Extract;

  procedure ShowJK(N: in String; M: in F3Mode; NewLine: in Boolean := True) is
  begin
    if M.F2Err then
      Show0(N & " F2Err:",False);
    else
      Show0(N & " Coeff:",False);
    end if;
    Show0(" I=" & Strng(M.I,2) & " J=" & Strng(M.J,3) & " K=" & Strng(M.K,3),NewLine);
  end ShowJK;

  procedure ShowType(M: in F3Mode) is
  begin
    Show1("Simple = ",M.Simple);
  end ShowType;

  procedure Show(M: in F3Mode) is
  begin
    ShowType(M);
    Show1(" F2Err = ",M.F2Err);
    Show1("     I =",M.I);
    Show1("     J =",M.J);
    Show1("     K =",M.K);
  end Show;

  function SameType(M1,M2: F3Mode) return Boolean is
  begin
    return (M1.Simple=M2.Simple);
  end SameType;

  procedure Show(M: in F3Modes) is
    F: constant Positive := M'First;
    MF: F3Mode renames M(F);
    OK: Boolean;
  begin
    if F /= 1 then raise Inconsistent_Data; end if;
    ShowType(MF);
    ShowJK("   1",MF);
    for K in 2 .. M'Last loop
      OK := SameType(MF,M(K));
      ShowJK(Strng(K,4),M(K),OK);
      if not OK then Show0("!!! wrong type"); end if;
    end loop;
  end Show;

  function Make(Siz: Positive; Simple,Trunc: Boolean) return F3Modes is
    Exclude_Errs: constant Boolean := Trunc or STrunc;
    Include_Errs: constant Boolean := not Exclude_Errs;
    S: constant Boolean := Simple or Exclude_Errs;
    Size2: constant Positive  := Siz/2;
    Size: constant Positive  := 2*Size2;     --- even
    Size1: constant Positive  := Size+1;
    Size32: constant Positive := 3*Size2;

    procedure Build(M: out F3Modes; Count: out Natural) is

      function Valid(T: F3Mode) return Boolean is
        I: Indx   renames T.I;
        J: FreqE1 renames T.J;
        K: FreqE2 renames T.K;
      begin
        if Include_Errs and I<0 then return False; end if;
        if not T.F2Err then
          return (((I+K) mod 2)=0) and then (J+K<=Size32);
        elsif J=Size1 then
          return K<=Size2;
        elsif K=Size1 then
          return J<=Size2;
        else
          return (J+K)=(Size32+1);
        end if;
      end Valid;

      procedure Set_Mode(I: Indx; J,K: in Natural; F2Err: in Boolean) is
        Tmp: constant F3Mode := (I,J,K,F2Err,S);
      begin
        if Valid(Tmp) then
          Count := Count+1;
          if Count <= M'Last then M(Count) := Tmp; end if;
        end if;
      end Set_Mode;

      procedure JK(I: in Integer) is
      begin
        for J in 1 .. Size loop
          for K in 1 .. Size loop
            Set_Mode(I,J,K,False);      --- coefficient
            if I>0 then Set_Mode(-I,J,K,False); end if;
          end loop;
        end loop;
      end JK;

    begin
      Count := 0;
      if NNOdd then
        if NN=1 then JK(0); end if;
        JK(1);
      elsif NN=0 then
        JK(0);
      else
        JK(2);
      end if;
      if Include_Errs and NN<3 then
        for J in 1 .. Size1 loop
          for K in 1 .. Size1 loop
            if NN<2 then
              for I in NonnegIndx loop
                Set_Mode(I,J,K,True);          -------  Fourier3 errors
              end loop;
            else
              if NNEven then
                Set_Mode(2,J,K,True);
              else
                Set_Mode(1,J,K,True);
              end if;
            end if;
          end loop;
        end loop;
      end if;
    end Build;

    Count: Natural;
  begin
    declare
      M: F3Modes(1..1);
    begin
      Build(M,Count);
    end;
    declare
      M: F3Modes(1..Count);
    begin
      Build(M,Count);
      return M;
    end;
  end Make;

end Hopf;
