with Globals, Protected_Counters, Ints, Strings;
use Globals, Protected_Counters, Ints, Strings;

pragma Elaborate_All (Globals,Protected_Counters);

package body Linear is

  use SM,SV;

  procedure Norm(Q: in Modes; V: in SVector; S: in out Scalar) is
    T,Tmp: Scalar;
  begin
    SetZero(S);
    for I in V'Range loop
      Norm(Q(I),T);
      AddProd(T,V(I),S,Tmp);
    end loop;
  end Norm;

  procedure AddProd(Q: in Modes; V: in SVector; F: in out Fun) is
  begin
    for I in V'Range loop
      AddProd(Q(I),V(I),F,Check_Modes);
    end loop;
  end AddProd;

  procedure Extract(Q: in Modes; F: in out Fun; V: in out SVector) is
  begin
    for I in Q'Range loop
      Extract(Q(I),F,V(I),Check_Modes);
    end loop;
  end Extract;

  procedure Op_Norm(Q: in Modes; S: out Radius) is
    --- modified for special norm
    --- deterministic: tasks working on independent data, and PMax is exact
    --Small: constant Flt := Flt(7)/Flt(8);
    Task_Error: Boolean := False;
    PC: Protected_Counter(Q'First);

    protected type PMax is procedure Max(T: in Radius); end PMax;
    protected body PMax is procedure Max(T: in Radius) is begin S := RMax(T,S); end Max; end PMax;
    PM: PMax;

    procedure Column_Loop is
      J: Integer;
      S2,T: Radius;
      S1: Scalar;
      F,G1,G2: Fun;
      use Flt_EF;
    begin
      Proper_Rounding;
      loop
        PC.Next(J);
        exit when J>Q'Last;
        if Verbosity>1 and then (J mod 1000) = 0 then Show1("j",J); end if;
        Assign(Q(J),F);
        Norm(Q(J),S1);
        LinOp(F,G1);
        if Need2(Q(J)) then
          Assign2(Q(J),F);
          LinOp(F,G2);
          S2 := MaxNorm(G1,G2);
        else
          S2 := MaxNorm(G1);
        end if;
        T := S2/Inf(S1);
        PM.Max(T);
        if Verbosity>2 or else T>One then ---Small then
          Show(Q(J));
          Show1("ratio:",T);
        end if;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Column_Loop;

    task type Column_Task_Type is pragma Storage_Size (Stack_Size); end Column_Task_Type;
    task body Column_Task_Type is begin Column_Loop; end Column_Task_Type;

    NT: constant Natural := Reserve_Tasks(Q'Last-Q'First,Op_Norm_Parallel);
  begin
    S := Zero;
    if NT=0 then
      Column_Loop;
    else
      declare
        Column_Task: array(1 .. NT) of Column_Task_Type;
        pragma Warnings (Off,Column_Task);
      begin
        Column_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
  end Op_Norm;

  procedure Enclosure(Q: in Modes; A: in out SMatrix) is
    --- deterministic: tasks working on independent data
    Task_Error: Boolean := False;
    PC: Protected_Counter(Q'First);

    procedure Column_Loop is
      J: Integer;
      V: SVector(Q'Range);
      F1,F2: Fun;
    begin
      Proper_Rounding;
      loop
        PC.Next(J);
        exit when J>Q'Last;
        if Verbosity>1 and then (J mod 1000)=0 then Show1("j",J); end if;
        Assign(Q(J),F1);
        LinOp(F1,F2);
        Extract(Q,F2,V);
        for I in Q'Range loop Copy(V(I),A(I,J)); end loop;
        if Verbosity>2 then Show(Q(J)); end if;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Column_Loop;

    task type Column_Task_Type is pragma Storage_Size (Stack_Size); end Column_Task_Type;
    task body Column_Task_Type is begin Column_Loop; end Column_Task_Type;

    NT: constant Natural := Reserve_Tasks(Q'Last-Q'First,Enclosure_Parallel);
  begin
    if Verbosity>1 and then Q'Last > 500 then Show1("Start Enclosure. Dim=",Q'Last); end if;
    if NT=0 then
      Column_Loop;
    else
      declare
        Column_Task: array(1 .. NT) of Column_Task_Type;
        pragma Warnings (Off,Column_Task);
      begin
        Column_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
    if Verbosity>1 then Show0("End Enclosure"); end if;
  end Enclosure;

  procedure Norm(A: in SMatrix; S: in out Scalar) is
    F1: constant Integer := A'First(1);
    L1: constant Integer := A'Last(1);
    SJ,T: Scalar;
  begin
    SetZero(S);
    for J in reverse A'Range(2) loop
      Norm(A(L1,J),SJ);
      for I in reverse F1 .. L1-1 loop
        Norm(A(I,J),T);
        Add(T,SJ);
      end loop;
      Max(SJ,S);
    end loop;
  end Norm;

  function Norm_Bound(A: SMatrix; B: Scalar) return Boolean is
    --- checks if Norm(A)<=B
    F1: constant Integer := A'First(1);
    L1: constant Integer := A'Last(1);
    SJ,T: Scalar;
  begin
    for J in reverse A'Range(2) loop
      Norm(A(L1,J),SJ);
      for I in reverse F1 .. L1-1 loop
        Norm(A(I,J),T);
        Add(T,SJ);
      end loop;
      if Compare(SJ,B)>0 then return False; end if;
    end loop;
    return True;
  end Norm_Bound;

  procedure OpNorm(Q: in Modes; A: in SMatrix; S: in out Scalar) is
    --- modified for special norm
    S1,S2,B11,B12,B22,U,V: Scalar;
    procedure Norm11(I,J: in Integer; T: in out Scalar) is
      --- 1d --> 1d
      A11: Scalar renames A(I,J);
    begin
      Norm(A11,T);
    end Norm11;
    procedure Norm21(I,J: in Integer; T: in out Scalar) is
      --- 1d --> 2d
      A11: Scalar renames A(I,J);
      A21: Scalar renames A(I+1,J);
    begin
      Sqr(A11,S1);
      Sqr(A21,S2);
      Add(S1,S2);
      Sqrt(S2,T);
    end Norm21;
    procedure Norm12(I,J: in Integer; T: in out Scalar) is
      --- 2d --> 1d
      A11: Scalar renames A(I,J);
      A12: Scalar renames A(I,J+1);
    begin
      Sqr(A11,S1);
      Sqr(A12,S2);
      Add(S1,S2);
      Sqrt(S2,T);
    end Norm12;
    procedure Norm22(I,J: in Integer; T: in out Scalar) is
      --- 2d --> 2d
      A11: Scalar renames A(I,J);
      A12: Scalar renames A(I,J+1);
      A21: Scalar renames A(I+1,J);
      A22: Scalar renames A(I+1,J+1);
    begin
      Sqr(A11,S1);
      Sqr(A21,S2);
      Sum(S1,S2,B11);   -- B11=A11*A11+A21*A21
      Prod(A11,A12,S1);
      Prod(A21,A22,S2);
      Sum(S1,S2,B12);   -- B12=A11*A12+A21*A22
      Sqr(A12,S1);
      Sqr(A22,S2);
      Sum(S1,S2,B22);   -- B22=A12*A12+A22*A22

      Sum(B11,B22,U);
      Mult(Half,U);     -- U=(B11+B22)/2
      Diff(B11,B22,V);
      Mult(Half,V);     -- V=(B11-B12)/2

      Sqr(V,S1);
      Sqr(B12,S2);
      Sum(S1,S2,T);
      Sqrt(T,S2);       -- S2=Sqrt(V*V+B12*B21)
      Norm(U,S1);
      Add(S1,S2);       -- S2=Sqrt(V*V+B12*B21)+|U|
      Sqrt(S2,T);
    end Norm22;

    Range2D,SkipI,Domain2D,SkipJ: Boolean;
    T,SJ: Scalar;
    W: SVector(Q'Range);
  begin
    for K in Q'Range loop Norm(Q(K),W(K)); end loop;
    SetZero(S);
    SkipJ := False;
    for J in A'Range(2) loop
      SetZero(SJ);
      if SkipJ then
        SkipJ := False;
      else
        Domain2D := Need2(Q(J));
        SkipI := False;
        for I in A'Range(1) loop
          if SkipI then
            SkipI := False;
          else
            Range2D := Need2(Q(I));
            if Range2D then         --- range 2d
              if Domain2D then
                Norm22(I,J,T);      -- 2d --> 2d
              else
                Norm21(I,J,T);      -- 1d --> 2d
              end if;
            else                    --- range 1d
              if Domain2D then
                Norm12(I,J,T);      -- 2d --> 1d
              else
                Norm11(I,J,T);      -- 1d --> 1d
              end if;
            end if;
            AddProd(W(I),T,SJ);
            SkipI := Range2D;
          end if; -- SkipI
        end loop;
        Quot(SJ,W(J),T);
        Max(T,S);
        SkipJ := Domain2d;
      end if; -- SkipJ
    end loop;
  end OpNorm;

end Linear;
