with Strings, Globals, Protected_Counters;
use Strings, Globals, Protected_Counters;

pragma Elaborate_All (Strings,Globals,Protected_Counters);

package body NSP.Fouriers is

  --- Misc

  procedure EffDegs(C: in FCoeff; D1,D2: out Natural) is
    --- not optimal
  begin
    D1 := 0;
    D2 := 0;
    for J in reverse PosFreq1 loop
      for K in reverse PosFreq2 loop
        if not IsZero(C(J,K)) then
          if J>D1 then D1 := J; end if;
          if K>D2 then D2 := K; end if;
        end if;
      end loop;
    end loop;
  end EffDegs;

  procedure EffDegs(E: in FErr; D1,D2: out Natural) is
    --- not optimal
  begin
    D1 := 0;
    D2 := 0;
    for J in reverse PosFreqE1 loop
      for K in reverse PosFreqE2 loop
        if E(J,K) /= EZero then
          if J>D1 then D1 := J; end if;
          if K>D2 then D2 := K; end if;
        end if;
      end loop;
    end loop;
  end EffDegs;

  function IsZero(C: FCoeff) return Boolean is
  begin
    for J in Freq1 loop
      for K in Freq2 loop
        if not IsZero(C(J,K)) then return False; end if;
      end loop;
    end loop;
    return True;
  end IsZero;

  function IsSharp(C: FCoeff) return Boolean is
  begin
    if Not_STrunc then
      for J in Freq1 loop
        for K in Freq2 loop
          if not IsSharp(C(J,K)) then return False; end if;
        end loop;
      end loop;
    end if;
    return True;
  end IsSharp;

  function IsZero(E: FErr) return Boolean is
  begin
    for J in FreqE1 loop
      for K in FreqE2 loop
        if E(J,K)/=EZero then return False; end if;
      end loop;
    end loop;
    return True;
  end IsZero;

  --- basic

  function IsPoly(F: Fourier3) return Boolean is
  begin
    return STrunc or else IsZero(F.E);
  end IsPoly;

  function IsSharp(F: Fourier3) return Boolean is
  begin
    return STrunc or else (IsSharp(F.C) and then IsZero(F.E));
  end IsSharp;

  function IsZero(F: Fourier3) return Boolean is
  begin
    return IsZero(F.C) and then (STrunc or else IsZero(F.E));
  end IsZero;

  procedure SetZero(F: in out Fourier3) is
    C: FCoeff renames F.C;
  begin
    for J in Freq1 loop
      for K in Freq2 loop SetZero(C(J,K)); end loop;
    end loop;
    F.E := (others => (others => EZero));
  end SetZero;

  procedure Copy(F1: in Fourier3; F2: in out Fourier3) is
    C1: FCoeff renames F1.C;
    C2: FCoeff renames F2.C;
  begin
    for J in Freq1 loop
      for K in Freq2 loop Copy(C1(J,K),C2(J,K)); end loop;
    end loop;
    if Not_STrunc then F2.E := F1.E; end if;
  end Copy;

  procedure Swap(F1,F2: in out Fourier3) is
    C1: FCoeff renames F1.C;
    C2: FCoeff renames F2.C;
  begin
    for J in Freq1 loop
      for K in Freq2 loop Swap(C1(J,K),C2(J,K)); end loop;
    end loop;
    if Not_STrunc then
      declare
        E1: constant FErr := F1.E;
      begin
        F1.E := F2.E;
        F2.E := E1;
      end;
    end if;
  end Swap;

  --- sets

  procedure BallAt0(R: in Radius; F: in out Fourier3) is
    --- no |frequencies|>1
  begin
    SetZero(F);
    if Not_STrunc then AddBall(OnePlusTiny*R,F.E(1,1)); end if;
  end BallAt0;

  function BallAt0(R: Radius) return Fourier3 is
    --- no |frequencies|>1
    F: Fourier3;
  begin
    BallAt0(R,F);
    return F;
  end BallAt0;

  procedure AddBall(R: in Flt; F: in out Fourier3) is
    --- meant for NN=1
    --- no |frequencies|>1
  begin
    if not (STrunc or else R=Zero) then
      AddBall(OnePlusTiny*abs(R),F.E(1,1));
    end if;
  end AddBall;

  procedure AddBall(S: in Scalar; F: in out Fourier3) is
    --- meant for NN=1
    --- no |frequencies|>1
  begin
    if not (STrunc or else IsZero(S)) then
      AddBall(OnePlusTiny*MaxNorm(S),F.E(1,1));
    end if;
  end AddBall;

  --- addition and multiplication etc.

  procedure Neg(F: in out Fourier3) is
    C: FCoeff renames F.C;
  begin
    for J in Freq1 loop
      for K in Freq2 loop Neg(C(J,K)); end loop;
    end loop;
  end Neg;

  procedure Neg(F1: in Fourier3; F2: in out Fourier3) is
    C1: FCoeff renames F1.C;
    C2: FCoeff renames F2.C;
  begin
    for J in Freq1 loop
      for K in Freq2 loop Neg(C1(J,K),C2(J,K)); end loop;
    end loop;
    if Not_STrunc then F2.E := F1.E; end if;
  end Neg;

  function "-"(F: Fourier3) return Fourier3 is
    G: Fourier3;
  begin
    Neg(F,G);
    return G;
  end "-";

  procedure Add(F1: in Fourier3; F2: in out Fourier3) is
    C1: FCoeff renames F1.C;
    C2: FCoeff renames F2.C;
  begin
    for J in Freq1 loop
      for K in Freq2 loop Add(C1(J,K),C2(J,K)); end loop;
    end loop;
    if Not_STrunc then
      declare
        E1: FErr renames F1.E;
        E2: FErr renames F2.E;
      begin
        for J in FreqE1 loop
          for K in FreqE2 loop Add(E1(J,K),E2(J,K)); end loop;
        end loop;
      end;
    end if;
  end Add;

  procedure Sum(F1,F2: in Fourier3; F3: in out Fourier3) is
    C1: FCoeff renames F1.C;
    C2: FCoeff renames F2.C;
    C3: FCoeff renames F3.C;
  begin
    for J in Freq1 loop
      for K in Freq2 loop Sum(C1(J,K),C2(J,K),C3(J,K)); end loop;
    end loop;
    if Not_STrunc then
      declare
        E1: FErr renames F1.E;
        E2: FErr renames F2.E;
        E3: FErr renames F3.E;
      begin
        for J in FreqE1 loop
          for K in FreqE2 loop Sum(E1(J,K),E2(J,K),E3(J,K)); end loop;
        end loop;
      end;
    end if;
  end Sum;

  function "+"(F1,F2: Fourier3) return Fourier3 is
    F3: Fourier3;
  begin
    Sum(F1,F2,F3);
    return F3;
  end "+";

  procedure Sub(F1: in Fourier3; F2: in out Fourier3) is
    C1: FCoeff renames F1.C;
    C2: FCoeff renames F2.C;
  begin
    for J in Freq1 loop
      for K in Freq2 loop Sub(C1(J,K),C2(J,K)); end loop;
    end loop;
    if Not_STrunc then
      declare
        E1: FErr renames F1.E;
        E2: FErr renames F2.E;
      begin
        for J in FreqE1 loop
          for K in FreqE2 loop Add(E1(J,K),E2(J,K)); end loop;
        end loop;
      end;
    end if;
  end Sub;

  procedure Diff(F1,F2: in Fourier3; F3: in out Fourier3) is
    C1: FCoeff renames F1.C;
    C2: FCoeff renames F2.C;
    C3: FCoeff renames F3.C;
  begin
    for J in Freq1 loop
      for K in Freq2 loop Diff(C1(J,K),C2(J,K),C3(J,K)); end loop;
    end loop;
    if Not_STrunc then
      declare
        E1: FErr renames F1.E;
        E2: FErr renames F2.E;
        E3: FErr renames F3.E;
      begin
        for J in FreqE1 loop
          for K in FreqE2 loop Sum(E1(J,K),E2(J,K),E3(J,K)); end loop;
        end loop;
      end;
    end if;
  end Diff;

  function "-"(F1,F2: Fourier3) return Fourier3 is
    F3: Fourier3;
  begin
    Diff(F1,F2,F3);
    return F3;
  end "-";

  procedure Mult(R: in Flt; F: in out Fourier3) is
  begin
    if R=Zero then
      SetZero(F);
    else
      declare
        C: FCoeff renames F.C;
        Tmp: Scalar;
      begin
        for J in Freq1 loop
          for K in Freq2 loop Mult(R,C(J,K),Tmp); end loop;
        end loop;
      end;
      if Not_STrunc then
        declare
          E: FErr renames F.E;
          A: constant Radius := abs(R);
        begin
          for J in FreqE1 loop
            for K in FreqE2 loop Mult(A,E(J,K)); end loop;
          end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(R: in Flt; F1: in Fourier3; F2: in out Fourier3) is
    C1: FCoeff renames F1.C;
    C2: FCoeff renames F2.C;
  begin
    if R=Zero then
      SetZero(F2);
    else
      for J in Freq1 loop
        for K in Freq2 loop Prod(R,C1(J,K),C2(J,K)); end loop;
      end loop;
      if Not_STrunc then
        declare
          E1: FErr renames F1.E;
          E2: FErr renames F2.E;
          A: constant Radius := abs(R);
        begin
          for J in FreqE1 loop
            for K in FreqE2 loop Prod(A,E1(J,K),E2(J,K)); end loop;
          end loop;
        end;
      end if;
    end if;
  end Prod;

  function "*"(R: Flt; F: Fourier3) return Fourier3 is
    G: Fourier3;
  begin
    Prod(R,F,G);
    return G;
  end "*";

  procedure AddProd(R: in Flt; F1: in Fourier3; F2: in out Fourier3) is
  begin
    if R/=Zero then
      declare
        C1: FCoeff renames F1.C;
        C2: FCoeff renames F2.C;
        Tmp: Scalar;
      begin
        for J in Freq1 loop
          for K in Freq2 loop AddProd(R,C1(J,K),C2(J,K),Tmp); end loop;
        end loop;
      end;
      if Not_STrunc then
        declare
          E1: FErr renames F1.E;
          E2: FErr renames F2.E;
          A: constant Radius := abs(R);
        begin
          for J in FreqE1 loop
            for K in FreqE2 loop AddProd(A,E1(J,K),E2(J,K)); end loop;
          end loop;
        end;
      end if;
    end if;
  end AddProd;

  procedure Mult(S: in Scalar; F: in out Fourier3) is
    C: FCoeff renames F.C;
  begin
    if IsZero(S) then
      SetZero(F);
    else
      for J in Freq1 loop
        for K in Freq2 loop Mult(S,C(J,K)); end loop;
      end loop;
      if Not_STrunc then
        declare
          E: FErr renames F.E;
          A: constant Radius := MaxNorm(S);
        begin
          for J in FreqE1 loop
            for K in FreqE2 loop Mult(A,E(J,K)); end loop;
          end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(S: in Scalar; F1: in Fourier3; F2: in out Fourier3) is
    C1: FCoeff renames F1.C;
    C2: FCoeff renames F2.C;
  begin
    if IsZero(S) then
      SetZero(F2);
    else
      for J in Freq1 loop
        for K in Freq2 loop Prod(S,C1(J,K),C2(J,K)); end loop;
      end loop;
      if Not_STrunc then
        declare
          E1: FErr renames F1.E;
          E2: FErr renames F2.E;
          A: constant Radius := MaxNorm(S);
        begin
          for J in FreqE1 loop
            for K in FreqE2 loop Prod(A,E1(J,K),E2(J,K)); end loop;
          end loop;
        end;
      end if;
    end if;
  end Prod;

  function "*"(S: Scalar; F: Fourier3) return Fourier3 is
    G: Fourier3;
  begin
    Prod(S,F,G);
    return G;
  end "*";

  procedure AddProd(S: in Scalar; F1: in Fourier3; F2: in out Fourier3) is
    C1: FCoeff renames F1.C;
    C2: FCoeff renames F2.C;
  begin
    if not IsZero(S) then
      for J in Freq1 loop
        for K in Freq2 loop AddProd(S,C1(J,K),C2(J,K)); end loop;
      end loop;
      if Not_STrunc then
        declare
          E1: FErr renames F1.E;
          E2: FErr renames F2.E;
          A: constant Radius := MaxNorm(S);
        begin
          for J in FreqE1 loop
            for K in FreqE2 loop AddProd(A,E1(J,K),E2(J,K)); end loop;
          end loop;
        end;
      end if;
    end if;
  end AddProd;

  --- functions

  function MaxNorm(E: FErr) return Radius is
    R: Flt := Zero;
  begin
    for J in reverse PosFreqE1 loop
      for K in reverse PosFreqE2 loop R := R+MaxNorm(E(J,K)); end loop;
    end loop;
    return OnePlusTiny*R;
  end MaxNorm;

  function MaxNorm(F: Fourier3) return Radius is
    C: FCoeff renames F.C;
    R: Flt;
  begin
    if STrunc then R := Zero; else R := MaxNorm(F.E); end if;
    for J in reverse PosFreq1 loop
      for K in reverse PosFreq2 loop R := R+MaxNorm(C(J,K)); end loop;
    end loop;
    return OnePlusTiny*R;
  end MaxNorm;

  function MaxNorm(F1,F2: Fourier3) return Radius is
    M: constant Positive := 128; --- a power of 2
    SPi: constant Scalar := Pi;
    R: Radius := MaxNorm(F1);
    X,Y: Scalar;
    F: Fourier3;
  begin
    for K in 1 .. M-1 loop
      Prod(Flt(K)/Flt(M),SPi,X);
      Cos(X,Y);
      Prod(Y,F1,F);
      Sin(X,Y);
      AddProd(Y,F2,F);
      R := RMax(MaxNorm(F),R);
    end loop;
    return R*(Flt(-M)/(Inf(SPi)-Flt(M))); -- R*M/(M-Pi)
  end MaxNorm;

  procedure Norm(F: in Fourier3; S: in out Scalar) is
    C: FCoeff renames F.C;
    Tmp: Scalar;
  begin
    if STrunc then SetZero(S); else DiskAt0(MaxNorm(F.E),S); end if;
    for J in reverse PosFreq1 loop
      for K in reverse PosFreq2 loop
        Norm(C(J,K),Tmp);
        Add(Tmp,S);
      end loop;
    end loop;
    Mult(OnePlusTiny,S,Tmp);
  end Norm;

  function Norm(F: Fourier3) return Scalar is
    S: Scalar;
  begin
    Norm(F,S);
    return S;
  end Norm;

  --- conversion and i/o

  procedure Show1(N: in String; F: in Fourier3; Hide0: in Boolean := True) is
    C: FCoeff renames F.C;
    E: FErr renames F.E;
    W: constant Positive := 4;
  begin
    if IsZero(F) then
      Put(N & "  Zero_Fourier3");
    else
      for J in Freq1 loop
        for K in Freq2 loop
          if not IsZero(C(J,K)) then
            Show1(N & " C " & Strng(J,W) & Strng(K,W) & " ",C(J,K),Hide0);
          end if;
        end loop;
      end loop;
      if Not_STrunc then
        New_Line;
        for J in FreqE1 loop
          for K in FreqE2 loop
            if E(J,K) /= EZero then
              Show1(N & " E " & Strng(J,W) & Strng(K,W) & " ",E(J,K),Hide0);
            end if;
          end loop;
        end loop;
      end if;
    end if;
    New_Line;
  end Show1;

  procedure Put(F: in File_Type; H: in Fourier3; Decimal: in Boolean := False) is
    --- saving only coefficients -1,0,1 of NSPoly
    PZero: constant NSPoly := PolyZero;
    HC: FCoeff renames H.C;
    HE: FErr renames H.E;
    D1,D2: Integer;
  begin
    EffDegs(HC,D1,D2);
    Put(F,D1);
    Put(F,D2);
    for J in 0 .. D1 loop
      for K in 0 .. D2 loop
        if J=0 or else K=0 then
          Put01(F,PZero,Decimal);  --- for compatibility
        else
          Put01(F,HC(J,K),Decimal);
        end if;
      end loop;
    end loop;
    if STrunc then
      Put(F,0);
      Put(F,0);
      Put01(F,EZero,Decimal);
    else
      EffDegs(HE,D1,D2);
      Put(F,D1);
      Put(F,D2);
      for J in 0 .. D1 loop
        for K in 0 .. D2 loop
          if J=0 or else K=0 then
            Put01(F,EZero,Decimal);  --- for compatibility
          else
            Put01(F,HE(J,K),Decimal);
          end if;
        end loop;
      end loop;
    end if;
  end Put;

  procedure AddCoeff01(M,N: in Natural; S: in NSPoly; F: in out Fourier3) is
  begin
    if not IsZero(S) then
      if (M <= Deg1) and then (N <= Deg2) then
        Add(S,F.C(M,N));
      elsif Not_STrunc then
        declare
          M0: constant Positive := IMin(M,DegE1);
          N0: constant Positive := IMin(N,DegE2);
        begin
          F.E(M0,N0) := F.E(M0,N0)+OnePlusTiny*MaxNorms1(S);
        end;
      end if;
    end if;
  end AddCoeff01;

  procedure Get(F: in File_Type; H: in out Fourier3; Decimal: in Boolean := False) is
    HE: FErr renames H.E;
    D1,D2: Integer := 0;
    J0,K0: Natural;
    E: NSErr := EZero;
    C: NSPoly;
  begin
    SetZero(H);
    Get(F,D1);
    Get(F,D2);
    for J in 0 .. D1 loop
      for K in 0 .. D2 loop
        Get01(F,C,Decimal);
        AddCoeff01(J,K,C,H);
      end loop;
    end loop;
    Get(F,D1);
    Get(F,D2);
    for J in 0 .. D1 loop
      J0 := IMin(IMax(J,1),DegE1);
      for K in 0 .. D2 loop
        K0 := IMin(IMax(K,1),DegE2);
        Get01(F,E,Decimal);
        Add(E,HE(J0,K0));
      end loop;
    end loop;
  end Get;

  procedure Write(Name: in String; H: in Fourier3; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>1 then Put_Line("Writing " & Name); end if;
    Create(F,Out_File,Name);
    Put(F,H,Decimal);
    Close(F);
  end Write;

  procedure Read(Name: in String; H: in out Fourier3; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>1 then Put_Line("Reading " & Name); end if;
    Open(F,In_File,Name);
    Get(F,H,Decimal);
    Close(F);
  end Read;

  --- misc

  procedure Show_Param_F3 is
  begin
    if STrunc then
      Show0("Scalar type is Numeric");
    else
      Show0("Scalar type is Ball");
    end if;
    if FProd_Parallel then
      Show0("FProd_Parallel is True");
    else
      Show0("FProd_Parallel is False");
    end if;
    Show1("Precision ",Get_Precision(SOne/Three));
    Show1("  NN =",NN);
    Show2("Degs =",Deg1,Deg2);
  end Show_Param_F3;

  procedure Free_Cache_F3 is
  begin
    Pool3.Free_All;
    Free_Cache(SHalf);
  end Free_Cache_F3;

  --------------- other procedures

  procedure Coeff(J,K: in Positive; F: in Fourier3; P: in out NSPoly) is
  begin
    if STrunc then
      Copy(F.C(J,K),P);
    else
      BallsAt0(F.E(J,K),P);
      Add(F.C(J,K),P);
    end if;
  end Coeff;

  procedure Coeff(I: in Integer; J,K: in Positive; F: in Fourier3; S: in out Scalar) is
  begin
    if STrunc then
      Coeff(I,F.C(J,K),S);
    else
      Coeff(I,F.C(J,K),F.E(J,K),S);
    end if;
  end Coeff;

  function Coeff(I: Integer; J,K: Positive; F: Fourier3) return Scalar is
    S: Scalar;
  begin
    Coeff(I,J,K,F,S);
    return S;
  end Coeff;

  procedure AddBall(J,K: in Natural; R: in Flt; F: in out Fourier3) is
    J0,K0: Natural;
  begin
    if not (STrunc or else R=Zero) then
      if J>DegE1 then J0 := DegE1; elsif J>0 then J0 := J; else J0 := 1; end if;
      if K>DegE1 then K0 := DegE1; elsif K>0 then K0 := K; else K0 := 1; end if;
      AddBall(OnePlusTiny*abs(R),F.E(J0,K0));
    end if;
  end AddBall;

  procedure AddBall(J,K: in Natural; S: in Scalar; F: in out Fourier3) is
    J0,K0: Natural;
  begin
    if not (STrunc or else IsZero(S)) then
      if J>DegE1 then J0 := DegE1; elsif J>0 then J0 := J; else J0 := 1; end if;
      if K>DegE1 then K0 := DegE1; elsif K>0 then K0 := K; else K0 := 1; end if;
      AddBall(OnePlusTiny*MaxNorm(S),F.E(J0,K0));
    end if;
  end AddBall;

  procedure Center(F: in out Fourier3) is
    C: FCoeff renames F.C;
  begin
    for J in Freq1 loop
      for K in Freq2 loop Center(C(J,K)); end loop;
    end loop;
    F.E := (others => (others => EZero));
  end Center;

  --- very NS specific

  function EE_LOP(E1,E2: FErr) return NSErr is
    --- norm factors via FourPlusTiny
    R: Radius;
    A1,B1,A2,B2: NSErr := EZero;
    C1,C2: NSErr;
  begin
    for J in reverse 1 .. DegE1 loop
      for K in reverse 1 .. DegE2 loop
        C1 := E1(J,K);
        C2 := E2(J,K);
        if C1 /= EZero then
          R := RFacs(J,K);
          A1 := A1+C1;
          B1 := B1+R*C1;
          if C2 /= EZero then
            A2 := A2+C2;
            B2 := B2+R*C2;
          end if;
        elsif C2 /= EZero then
          R := RFacs(J,K);
          A2 := A2+C2;
          B2 := B2+R*C2;
        end if;
      end loop;
    end loop;
    return FourPlusTiny*(A1*B2+B1*A2); --- the order matters
  end EE_LOP;

  procedure LOP(F1,F2: in Fourier3; F3: in out Fourier3; ZC1,ZE1,ZC2,ZE2: in Boolean) is
    --- SetZero(F3) first
    CC: constant Boolean := not (ZC1 or ZC2);
    CE: constant Boolean := not (ZC1 or ZE2);
    EC: constant Boolean := not (ZE1 or ZC2);
    EE: constant Boolean := not (ZE1 or ZE2);
    PCC,PCE,PEC: Protected_Counter(1);

    procedure CCProd_Loop(F: in out Fourier3) is
      --- norm factors via OnePlusTiny
      FC1: FCoeff renames F1.C;
      FC2: FCoeff renames F2.C;
      FC3: FCoeff renames F.C;
      FE3: FErr renames F.E;
      JDiffPos,KDiffPos: Boolean;
      J2,JJ1,JJ2,JSum,JDiff,KSum,KDiff,JK,KJ,NegLap1,NegLap2,NegLapDiff: Integer;
      Rm,Rp: NSErr;
      CDiv,Sm,Sp,Tmp: NSPoly;
    begin
      loop
        PCC.Next(J2);
        exit when J2>Deg1;
        JJ2 := J2*J2;
        for K2 in 1 .. Deg2 loop
          declare
            C2: NSPoly renames FC2(J2,K2);
          begin
            if not IsZero(C2) then
              NegLap2 := JJ2+K2*K2;
              for J1 in 1 .. Deg1 loop
                JJ1 := J1*J1;
                JK := J1*K2;
                JSum := J2+J1;
                JDiff := J2-J1;           --- notice J2-J1 (j-J in the notes)
                if JDiff<0 then
                  JDiffPos := False;
                  JDiff := -JDiff;
                else
                  JDiffPos := True;
                end if;
                for K1 in 1 .. Deg2 loop
                  declare
                    C1: NSPoly renames FC1(J1,K1);
                  begin
                    if not IsZero(C1) then
                      NegLap1 := JJ1+K1*K1;
                      NegLapDiff := NegLap2-NegLap1;
                      KJ := K1*J2;
                      KSum := K1+K2;
                      KDiff := K1-K2;       --- notice K1-K2 (K-k in the notes)
                      if KDiff<0 then
                        KDiffPos := False;
                        KDiff := -KDiff;
                      else
                        KDiffPos := True;
                      end if;
                      Prod(C1,C2,Tmp);                          --- FC1(J1,K1)*FC2(J2,K2)
                      Quot(Tmp,Flt(NegLap2*NegLap1),CDiv);      --- CDiv: divided by Q denominator
                      Prod(Flt(NegLapDiff*(JK-KJ)),CDiv,Sm);    --- Sm: multiplied by Q numerator (JK-KJ)
                      Prod(SFacs(JDiff,KDiff),Sm,Tmp);          --- (--)
                      if JDiffPos=KDiffPos then Add(Tmp,FC3(JDiff,KDiff)); else Sub(Tmp,FC3(JDiff,KDiff)); end if;

                      if JSum <= Deg1 and then KSum <= Deg2 then
                        Prod(SFacs(JSum,KSum),Sm,Tmp);          --- (++)
                        Add(Tmp,FC3(JSum,KSum));
                        Prod(Flt(NegLapDiff*(JK+KJ)),CDiv,Sp);  --- Sp: multiplied by Q numerator, (JK+KJ)
                        Prod(SFacs(JDiff,KSum),Sp,Tmp);         --- (-+)
                        if JDiffPos then Add(Tmp,FC3(JDiff,KSum)); else Sub(Tmp,FC3(JDiff,KSum)); end if;
                        Prod(SFacs(JSum,KDiff),Sp,Tmp);         --- (+-)
                        if KDiffPos then Add(Tmp,FC3(JSum,KDiff)); else Sub(Tmp,FC3(JSum,KDiff)); end if;
                      elsif Not_STrunc then
                        Rm := OnePlusTiny*MaxNorms(Sm);                                  --- Rm: multiplied by Q numerator (JK-KJ)
                        Rp := (OnePlusTiny*Flt(abs(NegLapDiff)*(JK+KJ)))*MaxNorms(CDiv); --- Rp: multiplied by Q numerator (JK+KJ)
                        FE3(JSum, KSum) := FE3(JSum, KSum)+RFacs(JSum, KSum)*Rm;         --- (++)
                        if JDiff/=0 then
                          FE3(JDiff,KSum) := FE3(JDiff,KSum)+RFacs(JDiff,KSum)*Rp;       --- (-+)
                        end if;
                        if KDiff/=0 then
                          FE3(JSum,KDiff) := FE3(JSum,KDiff)+RFacs(JSum,KDiff)*Rp;       --- (+-)
                        end if;
                      end if;
                    end if;
                  end;
                end loop;
              end loop;
            end if;
          end;
        end loop;
      end loop;
    end CCProd_Loop;

    procedure ECProd_Loop(FE3: in out FErr) is
      --- norm factors via OnePlusTiny
      FE1: FErr renames F1.E;
      FC2: FCoeff renames F2.C;
      J2,JSum,JDiff,KSum,KDiff: Integer;
      RFac2: Flt;
      E1,E2,E: NSErr;
    begin
      loop
        PEC.Next(J2);
        exit when J2>Deg1;
        for K2 in 1 .. Deg2 loop
          E2 := MaxNorms(FC2(J2,K2));                     --- right factor
          if E2 /= EZero then
            E2 := E2*OnePlusTiny;                         --- norm of FC2(J2,K2)
            RFac2 := RFacs(J2,K2);
            for J1 in 1 .. DegE1 loop
              JSum := IMin(J1+J2,DegE1);
              JDiff := IMax(J1-J2,1);
              for K1 in 1 .. DegE2 loop
                E1 := FE1(J1,K1);
                if E1 /= EZero then
                  KSum := IMin(K1+K2,DegE2);
                  KDiff := IMax(K1-K2,1);
                  E := (E1*E2)*RMax(RFacs(J1,K1),RFac2);  --- the order matters
                  FE3( JSum, KSum) := FE3( JSum, KSum)+E; --- (++)
                  FE3(JDiff, KSum) := FE3(JDiff, KSum)+E; --- (-+)
                  FE3( JSum,KDiff) := FE3( JSum,KDiff)+E; --- (+-)
                  FE3(JDiff,KDiff) := FE3(JDiff,KDiff)+E; --- (--)
                end if;
              end loop;
            end loop;
          end if;
        end loop;
      end loop;
    end ECProd_Loop;

    procedure CEProd_Loop(FE3: in out FErr) is
      --- same as above, except
      FE1: FErr renames F2.E;   --- order reversed
      FC2: FCoeff renames F1.C; --- order reversed
      J2,JSum,JDiff,KSum,KDiff: Integer;
      RFac2: Flt;
      E1,E2,E: NSErr;
    begin
      loop
        PCE.Next(J2);
        exit when J2>Deg1;
        for K2 in 1 .. Deg2 loop
          E2 := MaxNorms1(FC2(J2,K2));                    --- left factor
          if E2 /= EZero then
            E2 := E2*OnePlusTiny;                         --- norm of FC2(J2,K2)
            RFac2 := RFacs(J2,K2);
            for J1 in 1 .. DegE1 loop
              JSum := IMin(J1+J2,DegE1);
              JDiff := IMax(J1-J2,1);
              for K1 in 1 .. DegE2 loop
                E1 := FE1(J1,K1);
                if E1 /= EZero then
                  KSum := IMin(K1+K2,DegE2);
                  KDiff := IMax(K1-K2,1);
                  E := (E2*E1)*RMax(RFacs(J1,K1),RFac2);  --- the order matters
                  FE3( JSum, KSum) := FE3( JSum, KSum)+E; --- (++)
                  FE3(JDiff, KSum) := FE3(JDiff, KSum)+E; --- (-+)
                  FE3( JSum,KDiff) := FE3( JSum,KDiff)+E; --- (+-)
                  FE3(JDiff,KDiff) := FE3(JDiff,KDiff)+E; --- (--)
                end if;
              end loop;
            end loop;
          end if;
        end loop;
      end loop;
    end CEProd_Loop;

    protected type PF3_Type is procedure Add(F: in Fourier3); end PF3_Type;
    protected body PF3_Type is
      procedure Add(F: in Fourier3) is begin Add(F,F3); end Add;
    end PF3_Type;

    PF3: PF3_Type;

    procedure Prod_Loop is
      Tmp: LD_Pointer3;
    begin
      Pool3.Allocate(Tmp);
      declare
        F: Fourier3 renames Tmp.all.Data;
      begin
        SetZero(F);
        if CC then CCProd_Loop(F); end if;
        if EC then ECProd_Loop(F.E); end if;
        if CE then CEProd_Loop(F.E); end if;
        PF3.Add(F);
      end;
      Pool3.Recycle(Tmp);
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    task type Prod_Task_Type is end Prod_Task_Type;
    task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;

    NT: constant Natural := Reserve_Tasks(Deg1,FProd_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Exception in LOP"; end if;
    end if;
    if EE then F3.E(1,1) := F3.E(1,1)+EE_LOP(F1.E,F2.E); end if;
  end LOP;

  procedure Prod(F1,F2: in Fourier3; F3: in out Fourier3) is
    ZC1,ZC2,ZE1,ZE2: Boolean;
  begin
    SetZero(F3);
    ------------------
    ZC1 := IsZero(F1.C);
    ZE1 := STrunc or else IsZero(F1.E);
    if (ZC1 and then ZE1) then return; end if;
    ZC2 := IsZero(F2.C);
    ZE2 := STrunc or else IsZero(F2.E);
    if (ZC2 and then ZE2) then return; end if;
    LOP(F1,F2,F3,ZC1,ZE1,ZC2,ZE2);
    CleanOdd(F3);
  end Prod;

  procedure CleanOdd(F: in out Fourier3) is
    FC: FCoeff renames F.C;
    FE: FErr renames F.E;
  begin
    for J in Freq1 loop SetZero(FC(J,0)); end loop;
    for K in Freq2 loop SetZero(FC(0,K)); end loop;
    if Not_STrunc then
      --- redundant
      for J in FreqE1 loop FE(J,0) := EZero; end loop;
      for K in FreqE2 loop FE(0,K) := EZero; end loop;
    end if;
  end CleanOdd;

  procedure InvLinear(A: in Scalar; F: in out Fourier3) is
    -- |Lap|^{1/2}/(-Lap + A*D_t) = |Lap|^{1/2}(-Lap - A*D_t)/(Lap^2 - (A*D_t)^2)
    use Flt_EF;
    FC: FCoeff renames F.C;
    FE: FErr renames F.E;
    JJ,N: Integer;
    FL2,NegA,InvFL1,R2,InvR1: Flt;
    SL1,SL2,SL4,NA1,NA2,T,S,Sjk,Sn,PCos,PSin: Scalar;
  begin
    for J in PosFreq1 loop
      JJ := J*J;
      for K in PosFreq2 loop
        FL2 := Flt(JJ+K*K);
        Assign(FL2,SL2);           -- SL2 = J*J+K*K
        Assign(FL2*FL2,SL4);       -- SL4 = (J*J+K*K)**2
        Sqrt(SL2,SL1);             -- SL1 = Sqrt(J*J+K*K)
        declare
          P: NSPoly renames FC(J,K);
        begin
          if NNOffset=0 then       -- N=0
            Div(SL1,P(0));
          else
            SetZero(P(0));
          end if;
          for I in 1 .. ILast loop -- N>0
            Copy(P( I),PCos);
            Copy(P(-I),PSin);
            if not (IsZero(PCos) and then IsZero(PSin)) then
              N := I+NNOffset;
              Prod(Flt(N),A,NA1);    -- NA1 = N*A
              Sqr(NA1,NA2);          -- NA2 = (N*A)**2
              Sum(SL4,NA2,T);
              Quot(SL1,T,S);         -- S = Sqrt(J*J+K*K)/[(J*J+K*K)**2+(N*A)**2]
              Prod(SL2,S,Sjk);       -- Sjk = (J*J+K*K)*S
              Prod(NA1,S,Sn);        -- Sn = (N*A)*S

              Prod(Sjk,PCos,S);
              Prod(Sn,PSin,T);
              Diff(S,T,P(I));        -- P( I) = Sjk*PCos-Sn*PSin

              Prod(Sjk,PSin,S);
              Prod(Sn,PCos,T);
              Sum(S,T,P(-I));        -- P(-I) = Sjk*PSin+Sn*PCos
            end if;
          end loop;
        end;
      end loop;
    end loop;
    if Not_STrunc then
      NegA := -Inf(A);
      for J in PosFreqE1 loop
        JJ := J*J;
        for K in PosFreqE2 loop
          FL2 := Flt(JJ+K*K);
          InvFL1 := Sqrt(FL2)/FL2;              -- 1/Sqrt(J*J*K*K)
          declare
            E: NSErr renames FE(J,K);
          begin
            if NNOffset=0 then                  -- N=0
              E(0) := InvFL1*E(0);
            else
              E(0) := Zero;
            end if;
            for I in 1 .. ILast loop            -- N>0
              if E(I)/=Zero then
                N := I+NNOffset;
                R2 := -(Flt(2*N)*NegA);         -- 2*N*A         lower bound
                InvR1 := Sqrt(R2)/R2;           -- 1/Sqrt(2*N*A)
                E(I) := RMin(InvR1,InvFL1)*E(I);
              end if;
            end loop;
          end;
        end loop;
      end loop;
    end if;
    CleanOdd(F);
  end InvLinear;

  procedure InvLinear(A: in Scalar; F1: in Fourier3; F2: in out Fourier3) is
  begin
    Copy(F1,F2);
    InvLinear(A,F2);
  end InvLinear;

  procedure DtInvLinear(A: in Scalar; F: in out Fourier3) is
    -- D_t/(-Lap + A*D_t) = D_t (-Lap - A*D_t)/(Lap^2 - (A*D_t)^2)
  begin
    if NN=1 then
      declare
        FC: FCoeff renames F.C;
        FE: FErr renames F.E;
        JJ: Integer;
        FL2,R,NegA2: Flt;
        SL2,SL4,T,S,Sjk,Sn,PCos,PSin,A2: Scalar;
        use Flt_EF;
      begin
        Sqr(A,A2);                     -- A2 = A**2
        for J in PosFreq1 loop
          JJ := J*J;
          for K in PosFreq2 loop
            FL2 := Flt(JJ+K*K);
            Assign(FL2,SL2);           -- SL2 = J*J+K*K
            Assign(FL2*FL2,SL4);       -- SL4 = (J*J+K*K)**2
            declare
              P: NSPoly renames FC(J,K);
            begin
              SetZero(P(0));
              Copy(P( 1),PCos);
              Copy(P(-1),PSin);
              if not (IsZero(PCos) and then IsZero(PSin)) then
                Sum(SL4,A2,T);
                Inv(T,S);                -- S = 1/[(J*J+K*K)**2+A**2]
                Prod(SL2,S,Sjk);         -- Sjk = (J*J+K*K)*S
                Prod(A,S,Sn);            -- Sn = A*S

                Prod(Sn,PCos,T);
                Prod(Sjk,PSin,S);
                Sum(T,S,P(1));           -- P( 1) = Sn*PCos+Sjk*PSin

                Prod(Sn,PSin,T);
                Prod(Sjk,PCos,S);
                Diff(T,S,P(-1));         -- P(-1) = Sn*PSin-Sjk*PCos
              end if;
            end;
          end loop;
        end loop;
        if Not_STrunc then
          R := Inf(A);
          NegA2 := R*(-R);               -- -A**2
          for J in PosFreqE1 loop
            JJ := J*J;
            for K in PosFreqE2 loop
              FL2 := Flt(JJ+K*K);        -- FL2 = J*J+K*K
              declare
                E: NSErr renames FE(J,K);
              begin
                E(0) := Zero;
                if E(1)/=Zero then
                  R := -(NegA2-FL2*FL2); -- (J*J+K*K)**2+A**2  lower bound
                  E(1) := E(1)*Sqrt(R)/R;
                end if;
              end;
            end loop;
          end loop;
        end if;
      end;
      CleanOdd(F);
    else
      SetZero(F);
    end if;
  end DtInvLinear;

  procedure DtInvLinear(A: in Scalar; F1: in Fourier3; F2: in out Fourier3) is
  begin
    Copy(F1,F2);
    DtInvLinear(A,F2);
  end DtInvLinear;

  ------------------------------------------

  procedure Make_Force(F: in out Fourier3) is
  begin
    SetZero(F);
    Assign(Flt(-5),F.C(1,2));
    Assign(Flt(13),F.C(3,2));
  end Make_Force;

  function Make_Force return Fourier3 is
    F: Fourier3;
  begin
    Make_Force(F);
    return F;
  end Make_Force;

  procedure Make_NegInvLapForce(F: in out Fourier3) is
  begin
    SetZero(F);
    Assign(Flt(-1),F.C(1,2));
    Assign(Flt( 1),F.C(3,2));
  end Make_NegInvLapForce;

  function Make_NegInvLapForce return Fourier3 is
    F: Fourier3;
  begin
    Make_NegInvLapForce(F);
    return F;
  end Make_NegInvLapForce;

end NSP.Fouriers;
