with Recycling, NSP.Factors;

pragma Elaborate_All (Recycling,NSP.Factors);

generic

  Deg1,Deg2: in Positive;

package NSP.Fouriers is   --- NS specific real analytic Fourier series in 3 variables

  DegE1: constant Positive := 2*Deg1;
  DegE2: constant Positive := 2*Deg2;

  subtype Freq1  is Integer range 0 .. Deg1;
  subtype Freq2  is Integer range 0 .. Deg2;
  subtype FreqE1 is Integer range 0 .. DegE1;
  subtype FreqE2 is Integer range 0 .. DegE2;
  subtype PosFreq1  is Integer range 1 .. Deg1;
  subtype PosFreq2  is Integer range 1 .. Deg2;
  subtype PosFreqE1 is Integer range 1 .. DegE1;
  subtype PosFreqE2 is Integer range 1 .. DegE2;

  type FCoeff is array(Freq1,Freq2) of NSPoly;
  type FErr is array(FreqE1,FreqE2) of NSErr;

  type Fourier3 is
    record
      C: FCoeff;
      E: FErr;
    end record;

  --- basic
  function IsPoly(F: Fourier3) return Boolean;
  function IsSharp(F: Fourier3) return Boolean;
  function IsZero(F: Fourier3) return Boolean;
  procedure SetZero(F: in out Fourier3);
  procedure Copy(F1: in Fourier3; F2: in out Fourier3);
  procedure Swap(F1,F2: in out Fourier3);

  --- sets
  procedure BallAt0(R: in Radius; F: in out Fourier3);
  function BallAt0(R: Radius) return Fourier3;
  procedure AddBall(R: in Flt; F: in out Fourier3);
  procedure AddBall(S: in Scalar; F: in out Fourier3);

  --- addition and multiplication etc.
  procedure Neg(F: in out Fourier3);
  procedure Neg(F1: in Fourier3; F2: in out Fourier3);
  function "-"(F: Fourier3) return Fourier3;
  procedure Add(F1: in Fourier3; F2: in out Fourier3);
  procedure Sum(F1,F2: in Fourier3; F3: in out Fourier3);
  function "+"(F1,F2: Fourier3) return Fourier3;
  procedure Sub(F1: in Fourier3; F2: in out Fourier3);
  procedure Diff(F1,F2: in Fourier3; F3: in out Fourier3);
  function "-"(F1,F2: Fourier3) return Fourier3;
  procedure Mult(R: in Flt; F: in out Fourier3);
  procedure Prod(R: in Flt; F1: in Fourier3; F2: in out Fourier3);
  function "*"(R: Flt; F: Fourier3) return Fourier3;
  procedure AddProd(R: in Flt; F1: in Fourier3; F2: in out Fourier3);
  procedure Mult(S: in Scalar; F: in out Fourier3);
  procedure Prod(S: in Scalar; F1: in Fourier3; F2: in out Fourier3);
  function "*"(S: Scalar; F: Fourier3) return Fourier3;
  procedure AddProd(S: in Scalar; F1: in Fourier3; F2: in out Fourier3);

  --- functions
  function MaxNorm(F: Fourier3) return Radius;
  function MaxNorm(F1,F2: Fourier3) return Radius;
  procedure Norm(F: in Fourier3; S: in out Scalar);     --- uses DiskAt0
  function Norm(F: Fourier3) return Scalar;             --- uses DiskAt0

  --- conversion and i/o
  procedure Show1(N: in String; F: in Fourier3; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; H: in Fourier3; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; H: in out Fourier3; Decimal: in Boolean := False);
  procedure Write(Name: in String; H: in Fourier3; Decimal: in Boolean := False);
  procedure Read(Name: in String; H: in out Fourier3; Decimal: in Boolean := False);

  --- misc
  procedure Show_Param_F3;
  procedure Free_Cache_F3;

  --- other procedures
  procedure Coeff(J,K: in Positive; F: in Fourier3; P: in out NSPoly);
  procedure Coeff(I: in Integer; J,K: in Positive; F: in Fourier3; S: in out Scalar);
  function Coeff(I: Integer; J,K: Positive; F: Fourier3) return Scalar;
  procedure AddBall(J,K: in Natural; R: in Flt; F: in out Fourier3);
  procedure AddBall(J,K: in Natural; S: in Scalar; F: in out Fourier3);
  procedure Center(F: in out Fourier3);

  --- very NS specific
  procedure Prod(F1,F2: in Fourier3; F3: in out Fourier3); --- LOP
  procedure InvLinear(A: in Scalar; F: in out Fourier3);
  procedure InvLinear(A: in Scalar; F1: in Fourier3; F2: in out Fourier3);
  procedure DtInvLinear(A: in Scalar; F: in out Fourier3);
  procedure DtInvLinear(A: in Scalar; F1: in Fourier3; F2: in out Fourier3);
  procedure Make_Force(F: in out Fourier3);
  function Make_Force return Fourier3;
  procedure Make_NegInvLapForce(F: in out Fourier3);
  function Make_NegInvLapForce return Fourier3;

  package NSP_Factors is new NSP.Factors;
  use NSP_Factors;

  SFacs: constant Scalar_Array2 := MakeLapFacs(3*Deg1,3*Deg2);
  RFacs: constant Flt_Array2 := MaxNorms(SFacs);
  Tiny:  constant Radius := Flt'Scaling(One,Flt'Machine_Emin);
  OnePlusTiny: constant Radius := One+Tiny; --- norm factor
  FourPlusTiny: constant Radius := Four+Tiny;
  OneMinusTiny: constant Radius := One;     --- inverse norm factor

  FProd_Parallel: Boolean := False;
  Task_Error: Boolean := False;

private

  procedure EffDegs(C: in FCoeff; D1,D2: out Natural);
  procedure EffDegs(E: in FErr; D1,D2: out Natural);
  function IsZero(C: FCoeff) return Boolean;
  function IsSharp(C: FCoeff) return Boolean;
  function IsZero(E: FErr) return Boolean;
  function MaxNorm(E: FErr) return Radius;

  procedure AddCoeff01(M,N: in Natural; S: in NSPoly; F: in out Fourier3);
  function EE_LOP(E1,E2: FErr) return NSErr;
  procedure LOP(F1,F2: in Fourier3; F3: in out Fourier3; ZC1,ZE1,ZC2,ZE2: in Boolean);
  procedure CleanOdd(F: in out Fourier3);

  ----- recycling
  package Alloc3 is new Recycling (Item => Fourier3);
  subtype LD_Pointer3 is Alloc3.LD_Pointer;
  O3: Alloc3.Controlled_Protected_Data;
  Pool3: Alloc3.Protected_Data renames O3.P;

end NSP.Fouriers;
